<p>This is a reverse chronological list of the major <a href="https://tinybase.org/">TinyBase</a> releases, with highlighted features.</p><h2 id="v2-2">v2.2</h2><p>This release includes a new <a href="https://tinybase.org/api/tools"><code>tools</code></a> module. These tools are not intended for production use, but are instead to be used as part of your engineering workflow to perform tasks like generating APIs from schemas, or schemas from data. For example:</p>

```js
const store = createStore().setTable('pets', {
  fido: {species: 'dog'},
  felix: {species: 'cat'},
  cujo: {species: 'dog'},
});

const tools = createTools(store);
const [dTs, ts] = tools.getStoreApi('shop');
```

<p>This will generate two files:</p>

```js yolo
// -- shop.d.ts --
/* Represents the 'pets' Table. */
export type PetsTable = {[rowId: Id]: PetsRow};
/* Represents a Row when getting the content of the 'pets' Table. */
export type PetsRow = {species: string};
//...

// -- shop.ts --
export const createShop: typeof createShopDecl = () => {
  //...
};
```

<p>This release includes a new <code>tinybase</code> CLI tool which allows you to generate Typescript definition and implementation files direct from a schema file:</p>

```bash
npx tinybase getStoreApi schema.json shop api

    Definition: [...]/api/shop.d.ts
Implementation: [...]/api/shop.ts
```

<p>Finally, the <a href="https://tinybase.org/api/tools"><code>tools</code></a> module also provides ways to track the overall size and structure of a <a href="https://tinybase.org/api/store/interfaces/store/store"><code>Store</code></a> for use while debugging.</p><h2 id="v2-1">v2.1</h2><p>This release allows you to create indexes where a single <a href="https://tinybase.org/api/store/type-aliases/store/row"><code>Row</code></a> <a href="https://tinybase.org/api/common/type-aliases/identity/id"><code>Id</code></a> can exist in multiple slices. You can utilize this to build simple keyword searches, for example.</p><p>Simply provide a custom getSliceIdOrIds function in the <a href="https://tinybase.org/api/indexes/interfaces/indexes/indexes/methods/configuration/setindexdefinition"><code>setIndexDefinition</code></a> method that returns an array of <a href="https://tinybase.org/api/indexes/type-aliases/concept/slice"><code>Slice</code></a> <a href="https://tinybase.org/api/common/type-aliases/identity/ids"><code>Ids</code></a>, rather than a single <a href="https://tinybase.org/api/common/type-aliases/identity/id"><code>Id</code></a>:</p>

```js
store.setTable('pets', {
  fido: {species: 'dog'},
  felix: {species: 'cat'},
  rex: {species: 'dog'},
});

const indexes = createIndexes(store);
indexes.setIndexDefinition('containsLetter', 'pets', (_, rowId) =>
  rowId.split(''),
);

console.log(indexes.getSliceIds('containsLetter'));
// -> ['f', 'i', 'd', 'o', 'e', 'l', 'x', 'r']
console.log(indexes.getSliceRowIds('containsLetter', 'i'));
// -> ['fido', 'felix']
console.log(indexes.getSliceRowIds('containsLetter', 'x'));
// -> ['felix', 'rex']
```

<p>This functionality is showcased in the <a href="https://tinybase.org/demos/word-frequencies">Word Frequencies</a> demo if you would like to see it in action.</p><h2 id="v2-0">v2.0</h2><p><strong>Announcing the next major version of <a href="https://tinybase.org/">TinyBase</a> 2.0!</strong> This is an exciting release that evolves <a href="https://tinybase.org/">TinyBase</a> towards becoming a reactive, relational data store, complete with querying, sorting, and pagination. Here are a few of the highlights...</p><h3 id="query-engine">Query Engine</h3><p>The <a href="https://tinybase.org/guides/making-queries/using-queries/">flagship feature</a> of this release is the new <a href="https://tinybase.org/api/queries"><code>queries</code></a> module. This allows you to build expressive queries against your data with a SQL-adjacent API that we&#x27;ve cheekily called <a href="https://tinybase.org/guides/making-queries/tinyql/">TinyQL</a>. The query engine lets you select, join, filter, group, sort and paginate data. And of course, it&#x27;s all reactive!</p><p>The best way to see the power of this new engine is with the two new demos we&#x27;ve included this release:</p><p><img src="/car-analysis.webp" alt="Thumbnail of demo" title="Thumbnail of demo"> The <a href="https://tinybase.org/demos/car-analysis">Car Analysis</a> demo showcases the analytical query capabilities of <a href="https://tinybase.org/">TinyBase</a> v2.0, grouping and sorting dimensional data for lightweight analytical usage, graphing, and tabular display. <em><a href="https://tinybase.org/demos/car-analysis/">Try this demo here</a>.</em></p><p><img src="/movie-database.webp" alt="Thumbnail of demo" title="Thumbnail of demo"> The <a href="https://tinybase.org/demos/movie-database">Movie Database</a> demo showcases the relational query capabilities of <a href="https://tinybase.org/">TinyBase</a> v2.0, joining together information about movies, directors, and actors from across multiple source tables. <em><a href="https://tinybase.org/demos/movie-database/">Try this demo here</a>.</em></p><h3 id="sorting-and-pagination">Sorting and Pagination</h3><p>To complement the query engine, you can now sort and paginate <a href="https://tinybase.org/api/store/type-aliases/store/row"><code>Row</code></a> <a href="https://tinybase.org/api/common/type-aliases/identity/ids"><code>Ids</code></a>. This makes it very easy to build grid-like user interfaces (also shown in the demos above). To achieve this, the <a href="https://tinybase.org/api/store/interfaces/store/store"><code>Store</code></a> now includes the <a href="https://tinybase.org/api/store/interfaces/store/store/methods/getter/getsortedrowids"><code>getSortedRowIds</code></a> method (and the <a href="https://tinybase.org/api/store/interfaces/store/store/methods/listener/addsortedrowidslistener"><code>addSortedRowIdsListener</code></a> method for reactivity), and the <a href="https://tinybase.org/api/queries/interfaces/queries/queries"><code>Queries</code></a> object includes the equivalent <a href="https://tinybase.org/api/queries/interfaces/queries/queries/methods/result/getresultsortedrowids"><code>getResultSortedRowIds</code></a> method and <a href="https://tinybase.org/api/queries/interfaces/queries/queries/methods/listener/addresultsortedrowidslistener"><code>addResultSortedRowIdsListener</code></a> method.</p><p>These are also exposed in the optional <a href="https://tinybase.org/api/ui-react"><code>ui-react</code></a> module via the <a href="https://tinybase.org/api/ui-react/functions/store-hooks/usesortedrowids"><code>useSortedRowIds</code></a> hook, the <a href="https://tinybase.org/api/ui-react/functions/queries-hooks/useresultsortedrowids"><code>useResultSortedRowIds</code></a> hook, the <a href="https://tinybase.org/api/ui-react/functions/store-components/sortedtableview"><code>SortedTableView</code></a> component and the <a href="https://tinybase.org/api/ui-react/functions/queries-components/resultsortedtableview"><code>ResultSortedTableView</code></a> component, and so on.</p><h3 id="queries-in-the-ui-react-module"><a href="https://tinybase.org/api/queries/interfaces/queries/queries"><code>Queries</code></a> in the <a href="https://tinybase.org/api/ui-react"><code>ui-react</code></a> module</h3><p>The v2.0 query functionality is fully supported by the <a href="https://tinybase.org/api/ui-react"><code>ui-react</code></a> module (to match support for <a href="https://tinybase.org/api/store/interfaces/store/store"><code>Store</code></a>, <a href="https://tinybase.org/api/metrics/interfaces/metrics/metrics"><code>Metrics</code></a>, <a href="https://tinybase.org/api/indexes/interfaces/indexes/indexes"><code>Indexes</code></a>, and <a href="https://tinybase.org/api/relationships/type-aliases/concept/relationship"><code>Relationship</code></a> objects). The <a href="https://tinybase.org/api/ui-react/functions/queries-hooks/usecreatequeries"><code>useCreateQueries</code></a> hook memoizes the creation of app- or component-wide Query objects; and the <a href="https://tinybase.org/api/ui-react/functions/queries-hooks/useresulttable"><code>useResultTable</code></a> hook, <a href="https://tinybase.org/api/ui-react/functions/queries-hooks/useresultrow"><code>useResultRow</code></a> hook, <a href="https://tinybase.org/api/ui-react/functions/queries-hooks/useresultcell"><code>useResultCell</code></a> hook (and so on) let you bind you component to the results of a query.</p><p>This is, of course, supplemented with higher-level components: the <a href="https://tinybase.org/api/ui-react/functions/queries-components/resulttableview"><code>ResultTableView</code></a> component, the <a href="https://tinybase.org/api/ui-react/functions/queries-components/resultrowview"><code>ResultRowView</code></a> component, the <a href="https://tinybase.org/api/ui-react/functions/queries-components/resultcellview"><code>ResultCellView</code></a> component, and so on. See the <a href="https://tinybase.org/guides/making-queries/building-a-ui-with-queries">Building A UI With Queries</a> guide for more details.</p><h3 id="it-s-a-big-release">It&#x27;s a big release!</h3><p>Thank you for all your support as we brought this important new release to life, and we hope you enjoy using it as much as we did building it. Please provide feedback via <a href="https://github.com/tinyplex/tinybase">Github</a> and <a href="https://twitter.com/tinybasejs">Twitter</a>!</p><h2 id="v1-3-0">v1.3.0</h2><p>Adds support for explicit transaction start and finish methods, as well as listeners for transactions finishing.</p><p>The <a href="https://tinybase.org/api/store/interfaces/store/store/methods/transaction/starttransaction"><code>startTransaction</code></a> method and <a href="https://tinybase.org/api/store/interfaces/store/store/methods/transaction/finishtransaction"><code>finishTransaction</code></a> method allow you to explicitly enclose a transaction that will make multiple mutations to the <a href="https://tinybase.org/api/store/interfaces/store/store"><code>Store</code></a>, buffering all calls to the relevant listeners until it completes when you call the <a href="https://tinybase.org/api/store/interfaces/store/store/methods/transaction/finishtransaction"><code>finishTransaction</code></a> method.</p><p>Unlike the <a href="https://tinybase.org/api/store/interfaces/store/store/methods/transaction/transaction"><code>transaction</code></a> method, this approach is useful when you have a more &#x27;open-ended&#x27; transaction, such as one containing mutations triggered from other events that are asynchronous or not occurring inline to your code. You must remember to also call the <a href="https://tinybase.org/api/store/interfaces/store/store/methods/transaction/finishtransaction"><code>finishTransaction</code></a> method explicitly when the transaction is started with the <a href="https://tinybase.org/api/store/interfaces/store/store/methods/transaction/starttransaction"><code>startTransaction</code></a> method, of course.</p>

```js
store.setTables({pets: {fido: {species: 'dog'}}});
store.addRowListener('pets', 'fido', () => console.log('Fido changed'));

store.startTransaction();
store.setCell('pets', 'fido', 'color', 'brown');
store.setCell('pets', 'fido', 'sold', true);
store.finishTransaction();
// -> 'Fido changed'
```

<p>In addition, see the <a href="https://tinybase.org/api/store/interfaces/store/store/methods/listener/addwillfinishtransactionlistener"><code>addWillFinishTransactionListener</code></a> method and the <a href="https://tinybase.org/api/store/interfaces/store/store/methods/listener/adddidfinishtransactionlistener"><code>addDidFinishTransactionListener</code></a> method for details around listening to transactions completing.</p>

```js
store.addWillFinishTransactionListener((store, cellsTouched) =>
  console.log(`Cells touched: ${cellsTouched}`),
);

store.transaction(() => store.setCell('pets', 'fido', 'species', 'dog'));
// -> 'Cells touched: false'

store.transaction(() => store.setCell('pets', 'fido', 'color', 'walnut'));
// -> 'Cells touched: true'
// -> 'Fido changed'
```

<p>Together, this release allows stores to couple their transaction life-cycles together, which we need for the query engine.</p><h2 id="v1-2-0">v1.2.0</h2><p>This adds a way to revert transactions if they have not met certain conditions.</p><p>When using the <a href="https://tinybase.org/api/store/interfaces/store/store/methods/transaction/transaction"><code>transaction</code></a> method, you can provide an optional <code>doRollback</code> callback which should return true if you want to revert the whole transaction at its conclusion.</p><p>The callback is provided with two objects, <code>changedCells</code> and <code>invalidCells</code>, which list all the net changes and invalid attempts at changes that were made during the transaction. You will most likely use the contents of those objects to decide whether the transaction should be rolled back.</p><h2 id="v1-1-0">v1.1.0</h2><p>This release allows you to listen to invalid data being added to a <a href="https://tinybase.org/api/store/interfaces/store/store"><code>Store</code></a>, allowing you to gracefully handle errors, rather than them failing silently.</p><p>There is a new listener type <a href="https://tinybase.org/api/store/type-aliases/listener/invalidcelllistener"><code>InvalidCellListener</code></a> and a <a href="https://tinybase.org/api/store/interfaces/store/store/methods/listener/addinvalidcelllistener"><code>addInvalidCellListener</code></a> method in the <a href="https://tinybase.org/api/store/interfaces/store/store"><code>Store</code></a> interface.</p><p>These allow you to keep track of failed attempts to update the <a href="https://tinybase.org/api/store/interfaces/store/store"><code>Store</code></a> with invalid <a href="https://tinybase.org/api/store/type-aliases/store/cell"><code>Cell</code></a> data. These listeners can also be mutators, allowing you to address any failed writes programmatically.</p><p>For more information, please see the <a href="https://tinybase.org/api/store/interfaces/store/store/methods/listener/addinvalidcelllistener"><code>addInvalidCellListener</code></a> method documentation. In particular, this explains how this listener behaves for a <a href="https://tinybase.org/api/store/interfaces/store/store"><code>Store</code></a> with a <a href="https://tinybase.org/api/store/type-aliases/schema/schema"><code>Schema</code></a>.</p>